#include "pichart.h"
#include <complex.h>

typedef double complex Complex;

#define N 4194304
//#define N 32768
static Complex *restrict x,*restrict y,*restrict wn;
static int ncpu,ndepth;

static void fftserial(int n,int s,Complex *restrict x,Complex *restrict y){
    int m=n/2;
    if(m==1){
        y[0]=x[0];
        y[m]=x[s];
    } else {
        fftserial(m,s*2,x,y);
        fftserial(m,s*2,x+s,y+m);
    }
    for(int i=0;i<m;i++) {
        Complex z0=y[i],z1=wn[i*s]*y[i+m];
        y[i]=z0+z1;
        y[i+m]=z0-z1;
    }

}
static void fftblock(int imin,int imax,int m,int s,Complex *restrict y){
    for(int i=imin;i<imax;i++) {
        Complex z0=y[i],z1=wn[i*s]*y[i+m];
        y[i]=z0+z1;
        y[i+m]=z0-z1;
    }
}
static void fftparallel(int n,int s,Complex *restrict x,Complex *restrict y){
    if(s>ndepth||n<1024) {
        fftserial(n,s,x,y);
        return;
    }
    int m=n/2;
    spawn(,) fftparallel(m,s*2,x,y);
    fftparallel(m,s*2,x+s,y+m);
    sync;
    int cmax=ndepth/s+1;
    for(int c=0;c<cmax;c++){
        int imin=c*m/cmax,imax=(c+1)*m/cmax;
        spawn(,firstprivate(imin,imax)) fftblock(imin,imax,m,s,y);
    }
    sync;
}

double fourier(){
    x=mymalloc(N*sizeof(Complex),"x");
    y=mymalloc(N*sizeof(Complex),"y");
    wn=mymalloc(N/2*sizeof(Complex),"wn");
    double ops=2*N*log2(N)+3*N*log2(N);
    for(int i=0;i<N;i++) x[i]=sin(2*M_PI*i/N);
    double theta=-2*M_PI/N;
    for(int i=0;i<N/2;i++) wn[i]=cos(i*theta)+I*sin(i*theta);
    double tmin=1e32;
    int nmin=nprocs;
    for(ncpu=nprocs;ncpu>=1;ncpu=ncpu*FACTOR){
        workers(ncpu) {
            ndepth=ncpu<7?ncpu:ncpu*2;
            double ttot=0.0;
            for(int j=0;;j++){
                tic();
                if(ncpu==1) fftserial(N,1,x,y);
                else fftparallel(N,1,x,y);
                double t=toc();
                double e=cimag(y[N-1])-N/2;
                if(fabs(e)>0.5){
                    fprintf(stderr,"Error %25.15e is too big in FFT!\n",e);
                    exit(1);
                }
                if(t<tmin){ tmin=t; nmin=ncpu; }
                ttot+=t;
                if(ttot>5 && j>3) break;
            }
        }
    }
    printf("N=%d Workers=%d Sec=%g Mflops=%g",N,nmin,tmin,ops/tmin/1e6);
    if(fstress){
        printf("\n%21sPerforming %g minute stress test...","",fstress);
        fflush(stdout);
        ncpu=nmin;
        workers(ncpu) {
            ndepth=ncpu<7?ncpu:ncpu*2;
            tic();
            while(toc()<fstress*60){
                if(ncpu==1) fftserial(N,1,x,y);
                else fftparallel(N,1,x,y);
            }
        }
        printf("done");
    }
    free(wn); free(y); free(x);
    return ops/tmin/1e6;
}
