#include "pichart.h"

#define N 16777216

static int *x,*y,*xdata,*xgold;
static int ncpu,logcpu;
static void mergeserial(int m,int x[m],int n, int y[n],int z[n+m]){
    int k=0,i=0,j=0;
    while(i<m&&j<n){
        if(x[i]<=y[j]) z[k++]=x[i++];
        else z[k++]=y[j++];
    }
    while(i<m) z[k++]=x[i++];
    while(j<n) z[k++]=y[j++];
}
static void msortserial(int n,int p,int x[n],int y[n]){
    if(n==1) {
        if(p%2) y[0]=x[0];
        return;
    }
    int m=n/2;
    msortserial(m,p+1,x,y);
    msortserial(n-m,p+1,x+m,y+m);
    if(p%2) mergeserial(m,x,n-m,x+m,y);
    else mergeserial(m,y,n-m,y+m,x);
}
static int binsearch(int n,int y[n],int r){
    int a=0,b=n-1;
    if(y[a]>r) return 0;
    if(y[b]<r) return n;
    for(int m=(a+b)/2;a!=m;m=(a+b)/2){
        if(y[m]<=r) a=m;
        else b=m;
    }
    return b;
}
static void mergeparallel(int p,int m,int *x,int n, int *y,int z[n+m]){
    if(m<n){
        int k=m; m=n; n=k;
        int *t=x; x=y; y=t;
    }
    if(p>logcpu||m<4096){
        mergeserial(m,x,n,y,z);
        return;
    }
    int m2=m/2;
    int n2=binsearch(n,y,x[m2]);
    spawn(,) mergeparallel(p+1,m2,x,n2,y,z);
    mergeparallel(p+1,m-m2,x+m2,n-n2,y+n2,z+m2+n2);
    sync;
}
static void msortparallel(int n,int p,int x[n],int y[n]){
    if(p>=logcpu||n<1024) {
        msortserial(n,p,x,y);
        return;
    }
    int m=n/2;
    spawn(,) msortparallel(m,p+1,x,y);
    msortparallel(n-m,p+1,x+m,y+m);
    sync;
    if(p%2) mergeparallel(p,m,x,n-m,x+m,y);
    else mergeparallel(p,m,y,n-m,y+m,x);
}

double merge(){
    double ops=N*log2(N);
    x=mymalloc(N*sizeof(int),"x");
    y=mymalloc(N*sizeof(int),"y");
    xdata=mymalloc(N*sizeof(int),"xdata");
    xgold=mymalloc(N*sizeof(int),"xgold");
    srand(1234);
    for(int k=0;k<N;k++) xdata[k]=rand();
    memcpy(xgold,xdata,N*sizeof(int));
    msortserial(N,0,xgold,y);
    double tmin=1e32;
    int nmin=nprocs;
    for(ncpu=nprocs;ncpu>=1;ncpu=ncpu*FACTOR){
        workers(ncpu){
            logcpu=log2(ncpu)+2.99;
            double ttot=0.0;
            for(int j=0;;j++){
                memcpy(x,xdata,N*sizeof(int));
                tic();
                if(ncpu==1) msortserial(N,0,x,y);
                else msortparallel(N,0,x,y);
                double t=toc();
                for(int k=0;k<N;k++) if(x[k]!=xgold[k]) {
                    fprintf(stderr,"Error in parallel sort at index %d!\n",k);
                    exit(1);
                }
                if(t<tmin){ tmin=t; nmin=ncpu; }
                ttot+=t;
                if(ttot>5 && j>3) break;
            }
        }
    }
    printf("N=%d Workers=%d Sec=%g Mops=%g",N,nmin,tmin,ops/tmin/1e6);
    if(fstress){
        printf("\n%21sPerforming %g minute stress test...","",fstress);
        fflush(stdout);
        ncpu=nmin;
        workers(ncpu) {
            logcpu=log2(ncpu)+2.99;
            tic();
            while(toc()<fstress*60){
                memcpy(x,xdata,N*sizeof(int));
                if(ncpu==1) msortserial(N,0,x,y);
                else msortparallel(N,0,x,y);
            }
        }
        printf("done");
    }
    free(xgold); free(xdata); free(y); free(x);
    return ops/tmin/1e6;
}
