#ifndef _PICHART_H
#define RELEASE "40"

#include <sys/time.h>
#include <sys/resource.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>

#if defined _OPENMP
#  define VERSION "OPENMP version " RELEASE
#  define SVGNAME "pichart-openmp.svg"
#  include <omp.h>
#  define do_pragma(Z) _Pragma(#Z)
#  define spawn(X,Y) do_pragma(omp task default(shared) Y) X
#  define sync _Pragma("omp taskwait")
#  define get_ncpu(X) omp_get_num_procs(X)
#  define workers(X) \
    omp_set_dynamic(0); \
    if(X) omp_set_num_threads(X); \
    else omp_set_num_threads(1); \
    _Pragma("omp parallel default(shared)") \
    _Pragma("omp single")
#elif defined _CILKPLUS
#  define VERSION "CILKPLUS version " RELEASE
#  define SVGNAME "pichart-cilk.svg"
#  include <cilk/cilk.h>
#  include <cilk/cilk_api.h>
#  define spawn(X,Y) X cilk_spawn
#  define sync cilk_sync
#  define get_ncpu(X) __cilkrts_get_nworkers(X)
#  define workers(X) \
   __cilkrts_end_cilk(); \
   if(X){ \
      char buf[32]; \
      sprintf(buf,"%d",X); \
      int r=__cilkrts_set_param("nworkers",buf); \
      if(r!=__CILKRTS_SET_PARAM_SUCCESS){ \
          fprintf(stderr,"Error: unable to set nworkers to %d!\n",X); \
          exit(1); \
      } \
      __cilkrts_init(); \
   }
#elif defined _OPENCILK
#  define VERSION "OPENCILK version " RELEASE
#  define SVGNAME "pichart-opencilk.svg"
#  include <cilk/cilk.h>
#  include <cilk/cilk_api.h>
#  define spawn(X,Y) X cilk_spawn
#  define sync cilk_sync
#  define get_ncpu(X) __cilkrts_get_nworkers(X)
#  define workers(X) \
/*   __attribute__((destructor)) void __default_cilkrts_shutdown(); \
     __default_cilkrts_shutdown(); \
     if(X){ \
        void __cilkrts_internal_set_nworkers(unsigned int nworkers); \
        __cilkrts_internal_set_nworkers(X); \
        __attribute__((constructor)) void __default_cilkrts_startup(); \
        __default_cilkrts_startup(); \
     } */
#else
#  define _SERIAL
#  define VERSION "Serial version " RELEASE
#  define SVGNAME "pichart-serial.svg"
#  define spawn(X,Y) X
#  define sync
#  define get_ncpu(X) 1
#  define workers(X)
#endif

#define FACTOR 1/2

typedef struct { char *m; double t[4]; } datatype;
extern int nprocs;
extern double fstress;
extern void tic();
extern double toc();
extern void *mymalloc(size_t size,char *p);

extern double sieve();
extern double merge();
extern double fourier();
extern double lorenz();

#endif
