#include "pichart.h"

#define getbit(p,n) (p[n>>3]&(0x01<<(n&0x07)))
#define setbit(p,n) (p[n>>3]|=(0x01<<(n&0x07)))

#define Limit 268435456
//#define Slice 64
#define Slice 2048

static int cgold;
static char *notPrimebits;
static int C[Slice];
static int ncpu;

static int ops;
static int sievegold(int limit){
    ops=0;
    int imax=sqrt((double)limit)+0.5;
    int i,c;
    for(i=2,c=0;i<=imax;i++){
        ops++;
        if(!getbit(notPrimebits,i)){
            c++;
            int j;
            for(j=i*i;j<=limit;j+=i){
                ops++;
                setbit(notPrimebits,j);
            }
        }
    }
    for(i=imax;i<=limit;i++){
        ops++;
        if(!getbit(notPrimebits,i)) c++;
    }
    return c;
}
static int serialsieve(int limit){
    int imax=sqrt((double)limit)+0.5;
    int i,c;
    for(i=2,c=0;i<=imax;i++){
        if(!getbit(notPrimebits,i)){
            c++;
            int j;
            for(j=i*i;j<=limit;j+=i)
                setbit(notPrimebits,j);
        }
    }
    for(i=imax;i<=limit;i++){
        if(!getbit(notPrimebits,i)) c++;
    }
    return c;
}

static int setrange(int imax,int jmin, int jmax){
    int i,c;
    for(i=2;i<=imax;i++){
        if(!getbit(notPrimebits,i)){
            int j=jmin+i-1; j-=j%i;
            for(;j<=jmax;j+=i)
                setbit(notPrimebits,j);
        }
    }
    for(i=jmin,c=0;i<=jmax;i++){
        if(!getbit(notPrimebits,i)) c++;
    }
    return c;
}
static inline int round8(int j){
    j=j+7; j-=j%8;
    return j;
}
static int parallelsieve(int limit){
    int imax=round8(sqrt((double)limit)+1.5)-1;
    int c=serialsieve(imax);
    memset(C,0,sizeof(int)*Slice);
    for(int n=0;n<Slice;n++){
        int jmin,jmax;
        jmin=round8(imax+1+(int64_t)n*(Limit-imax)/Slice);
        if(n==Slice-1) jmax=limit;
        else jmax=round8(imax+1+(int64_t)(n+1)*(Limit-imax)/Slice)-1;
        spawn(C[n]=,firstprivate(n,imax,jmin,jmax))
            setrange(imax,jmin,jmax);
    }
    sync;
    for(int n=0;n<Slice;n++) c+=C[n];
    return c;
}

static int blocksieve(int limit){
    int imax=round8(sqrt((double)limit)+1.5)-1;
    int c=serialsieve(imax);
    memset(C,0,sizeof(int)*Slice);
    for(int n=0;n<Slice;n++){
        int jmin,jmax;
        jmin=round8(imax+1+(int64_t)n*(Limit-imax)/Slice);
        if(n==Slice-1) jmax=limit;
        else jmax=round8(imax+1+(int64_t)(n+1)*(Limit-imax)/Slice)-1;
        C[n]=setrange(imax,jmin,jmax);
    }
    for(int n=0;n<Slice;n++) c+=C[n];
    return c;
}

double sieve(){
    notPrimebits=mymalloc(Limit/8+1,"notPrimebits");
    memset(notPrimebits,0,Limit/8+1);
    cgold=sievegold(Limit);
    double tmin=1e32;
    int nmin=nprocs;
    for(ncpu=nprocs;ncpu>=1;ncpu=ncpu*FACTOR){
        workers(ncpu){
            double ttot=0.0;
            for(int j=0;;j++){
                memset(notPrimebits,0,Limit/8+1);
                tic();
                int c;
                if(ncpu==1) c=blocksieve(Limit);
                else c=parallelsieve(Limit);
                double t=toc();
                if(c!=cgold){
                    fprintf(stderr,"Wrong number %d of primes found!\n",c);
                    exit(1);
                }
                if(t<tmin){ tmin=t; nmin=ncpu; }
                ttot+=t;
                if(ttot>5 && j>3) break;
            }
        }
    }
    printf("P=%d Workers=%d Sec=%g Mops=%g",
        cgold,nmin,tmin,ops/tmin/1e6);
    if(fstress){
        printf("\n%21sPerforming %g minute stress test...","",fstress);
        fflush(stdout);
        ncpu=nmin;
        workers(ncpu) {
            tic();
            while(toc()<fstress*60){
                memset(notPrimebits,0,Limit/8+1);
                if(ncpu==1) blocksieve(Limit);
                else parallelsieve(Limit);
            }
        }
        printf("done");
    }
    free(notPrimebits);
    return ops/tmin/1e6;
}
